@@ -1,46 +1,35 @@
 package YAML::Tiny;
 
-# Current load overhead 132k
-
-use 5.005;
 use strict;
-
-use vars qw{$VERSION $errstr};
 BEGIN {
-	$VERSION = '1.00';
-	$errstr  = '';
+	require 5.004;
+	require Exporter;
+	$YAML::Tiny::VERSION   = '1.22_01';
+	$YAML::Tiny::errstr    = '';
+	@YAML::Tiny::ISA       = qw{ Exporter  };
+	@YAML::Tiny::EXPORT_OK = qw{
+		Load     Dump
+		LoadFile DumpFile
+		freeze   thaw
+		};
 }
 
-# Create the main error hash
-my %ERROR = (
-	YAML_PARSE_ERR_NO_FINAL_NEWLINE => "Stream does not end with newline character",
-	
-);
-
-my %NO = (
-	'%' => 'YAML::Tiny does not support directives',
-	'&' => 'YAML::Tiny does not support anchors',
-	'*' => 'YAML::Tiny does not support aliases',
-	'?' => 'YAML::Tiny does not support explicit mapping keys',
-	':' => 'YAML::Tiny does not support explicit mapping values',
-	'!' => 'YAML::Tiny does not support explicit tags',
-);
-
 my $ESCAPE_CHAR = '[\\x00-\\x08\\x0b-\\x0d\\x0e-\\x1f\"\n]';
 
 # Escapes for unprintable characters
-my @UNPRINTABLE = qw(z    x01  x02  x03  x04  x05  x06  a
-                     x08  t    n    v    f    r    x0e  x0f
-                     x10  x11  x12  x13  x14  x15  x16  x17
-                     x18  x19  x1a  e    x1c  x1d  x1e  x1f
-                    );
+my @UNPRINTABLE = qw(
+	z    x01  x02  x03  x04  x05  x06  a
+	x08  t    n    v    f    r    x0e  x0f
+	x10  x11  x12  x13  x14  x15  x16  x17
+	x18  x19  x1a  e    x1c  x1d  x1e  x1f
+);
 
 # Printable characters for escapes
 my %UNESCAPES = (
 	z => "\x00", a => "\x07", t    => "\x09",
 	n => "\x0a", v => "\x0b", f    => "\x0c",
 	r => "\x0d", e => "\x1b", '\\' => '\\',
-	);
+);
 
 # Create an empty YAML::Tiny object
 sub new {
@@ -60,9 +49,10 @@ sub read {
 
 	# Slurp in the file
 	local $/ = undef;
-	open CFG, $file or return $class->_error( "Failed to open file '$file': $!" );
+	local *CFG;
+	open( CFG, $file ) or return $class->_error( "Failed to open file '$file': $!" );
 	my $contents = <CFG>;
-	close CFG;
+	close( CFG ) or return $class->_error( "Failed to close file '$file': $!" );
 
 	$class->read_string( $contents );
 }
@@ -76,26 +66,26 @@ sub read_string {
 	return undef unless defined $_[0];
 	return $self unless length $_[0];
 	unless ( $_[0] =~ /[\012\015]+$/ ) {
-		return $class->_error('YAML_PARSE_ERR_NO_FINAL_NEWLINE');
+		return $class->_error("Stream does not end with newline character");
 	}
 
 	# Split the file into lines
-	my @lines = grep { ! /^\s*(?:\#.+)?$/ }
+	my @lines = grep { ! /^\s*(?:\#.*)?$/ }
 	            split /(?:\015{1,2}\012|\015|\012)/, shift;
 
 	# A nibbling parser
 	while ( @lines ) {
 		# Do we have a document header?
-		if ( $lines[0] =~ /^---(?:\s*(.+)\s*)?$/ ) {
+		if ( $lines[0] =~ /^---\s*(?:(.+)\s*)?$/ ) {
 			# Handle scalar documents
 			shift @lines;
-			if ( defined $1 ) {
+			if ( defined $1 and $1 !~ /^(?:\#.+|\%YAML:[\d\.]+)$/ ) {
 				push @$self, $self->_read_scalar( "$1", [ undef ], \@lines );
 				next;
 			}
 		}
 
-		if ( ! @lines or $lines[0] =~ /^---(?:\s*(.+)\s*)?$/ ) {
+		if ( ! @lines or $lines[0] =~ /^---\s*(?:(.+)\s*)?$/ ) {
 			# A naked document
 			push @$self, undef;
 
@@ -112,30 +102,31 @@ sub read_string {
 			$self->_read_hash( $document, [ length($1) ], \@lines );
 
 		} else {
-			die "CODE INCOMPLETE";
+			die "YAML::Tiny does not support the line '$lines[0]'";
 		}
 	}
 
 	$self;
 }
 
-sub _check_support {
-	# Check if we support the next char
-	my $errstr = $NO{substr($_[1], 0, 1)};
-	Carp::croak($errstr) if $errstr;
-}
-
 # Deparse a scalar string to the actual scalar
 sub _read_scalar {
 	my ($self, $string, $indent, $lines) = @_;
+
+	# Trim trailing whitespace
+	$string =~ s/\s*$//;
+
+	# Explitic null/undef
 	return undef if $string eq '~';
-	if ( $string =~ /^'(.*?)'$/ ) {
+
+	# Quotes
+	if ( $string =~ /^\'(.*?)\'$/ ) {
 		return '' unless defined $1;
 		my $rv = $1;
-		$rv =~ s/''/'/g;
+		$rv =~ s/\'\'/\'/g;
 		return $rv;
 	}
-	if ( $string =~ /^"((?:\\.|[^"])*)"$/ ) {
+	if ( $string =~ /^\"((?:\\.|[^\"])*)\"$/ ) {
 		my $str = $1;
 		$str =~ s/\\"/"/g;
 		$str =~ s/\\([never\\fartz]|x([0-9a-fA-F]{2}))/(length($1)>1)?pack("H2",$2):$UNESCAPES{$1}/gex;
@@ -145,10 +136,19 @@ sub _read_scalar {
 		# A quote with folding... we don't support that
 		die "YAML::Tiny does not support multi-line quoted scalars";
 	}
-	unless ( $string eq '>' or $string eq '|' ) {
-		# Regular unquoted string
-		return $string;
+
+	# Null hash and array
+	if ( $string eq '{}' ) {
+		# Null hash
+		return {};		
 	}
+	if ( $string eq '[]' ) {
+		# Null array
+		return [];
+	}
+
+	# Regular unquoted string
+	return $string unless $string =~ /^[>|]/;
 
 	# Error
 	die "Multi-line scalar content missing" unless @$lines;
@@ -168,7 +168,9 @@ sub _read_scalar {
 		push @multiline, substr(shift(@$lines), length($1));
 	}
 
-	join( ($string eq '>' ? ' ' : "\n"), @multiline ) . "\n";
+	my $j = (substr($string, 0, 1) eq '>') ? ' ' : "\n";
+	my $t = (substr($string, 1, 1) eq '-') ? '' : "\n";
+	return join( $j, @multiline ) . $t;
 }
 
 # Parse an array
@@ -176,6 +178,10 @@ sub _read_array {
 	my ($self, $array, $indent, $lines) = @_;
 
 	while ( @$lines ) {
+		# Check for a new document
+		return 1 if $lines->[0] =~ /^---\s*(?:(.+)\s*)?$/;
+
+		# Check the indent level
 		$lines->[0] =~ /^(\s*)/;
 		if ( length($1) < $indent->[-1] ) {
 			return 1;
@@ -183,7 +189,7 @@ sub _read_array {
 			die "Hash line over-indented";
 		}
 
-		if ( $lines->[0] =~ /^(\s*\-\s+)\S+\s*:(?:\s+|$)/ ) {
+		if ( $lines->[0] =~ /^(\s*\-\s+)[^\'\"]\S*\s*:(?:\s+|$)/ ) {
 			# Inline nested hash
 			my $indent2 = length("$1");
 			$lines->[0] =~ s/-/ /;
@@ -197,6 +203,10 @@ sub _read_array {
 
 		} elsif ( $lines->[0] =~ /^\s*\-\s*$/ ) {
 			shift @$lines;
+			unless ( @$lines ) {
+				push @$array, undef;
+				return 1;
+			}
 			if ( $lines->[0] =~ /^(\s*)\-/ ) {
 				my $indent2 = length("$1");
 				if ( $indent->[-1] == $indent2 ) {
@@ -213,11 +223,11 @@ sub _read_array {
 				$self->_read_hash( $array->[-1], [ @$indent, length("$1") ], $lines );
 
 			} else {
-				die "CODE INCOMPLETE";
+				die "YAML::Tiny does not support the line '$lines->[0]'";
 			}
 
 		} else {
-			die "CODE INCOMPLETE";
+			die "YAML::Tiny does not support the line '$lines->[0]'";
 		}
 	}
 
@@ -229,6 +239,10 @@ sub _read_hash {
 	my ($self, $hash, $indent, $lines) = @_;
 
 	while ( @$lines ) {
+		# Check for a new document
+		return 1 if $lines->[0] =~ /^---\s*(?:(.+)\s*)?$/;
+
+		# Check the indent level
 		$lines->[0] =~/^(\s*)/;
 		if ( length($1) < $indent->[-1] ) {
 			return 1;
@@ -237,7 +251,7 @@ sub _read_hash {
 		}
 
 		# Get the key
-		unless ( $lines->[0] =~ s/^\s*(\S+)\s*:(\s+|$)// ) {
+		unless ( $lines->[0] =~ s/^\s*([^\'\"][^\n]*?)\s*:(\s+|$)// ) {
 			die "Bad hash line";
 		}
 		my $key = $1;
@@ -249,12 +263,16 @@ sub _read_hash {
 		} else {
 			# An indent
 			shift @$lines;
+			unless ( @$lines ) {
+				$hash->{$key} = undef;
+				return 1;
+			}
 			if ( $lines->[0] =~ /^(\s*)-/ ) {
 				$hash->{$key} = [];
 				$self->_read_array( $hash->{$key}, [ @$indent, length($1) ], $lines );
 			} elsif ( $lines->[0] =~ /^(\s*)./ ) {
 				my $indent2 = length("$1");
-				if ( $indent->[-1] == $indent2 ) {
+				if ( $indent->[-1] >= $indent2 ) {
 					# Null hash entry
 					$hash->{$key} = undef;
 				} else {
@@ -281,6 +299,8 @@ sub write {
 		);
 	print CFG $self->write_string;
 	close CFG;
+
+	return 1;
 }
 
 # Save an object to a string
@@ -300,18 +320,26 @@ sub write_string {
 
 		# A scalar document
 		} elsif ( ! ref $cursor ) {
-			$lines[-1] .= $self->_write_scalar( $cursor );
+			$lines[-1] .= ' ' . $self->_write_scalar( $cursor );
 
 		# A list at the root
 		} elsif ( ref $cursor eq 'ARRAY' ) {
-			push @lines, $self->_write_array( $indent, $cursor );
+			unless ( @$cursor ) {
+				$lines[-1] .= ' []';
+				next;
+			}
+			push @lines, $self->_write_array( $cursor, $indent, {} );
 
 		# A hash at the root
 		} elsif ( ref $cursor eq 'HASH' ) {
-			push @lines, $self->_write_hash( $indent, $cursor );
+			unless ( %$cursor ) {
+				$lines[-1] .= ' {}';
+				next;
+			}
+			push @lines, $self->_write_hash( $cursor, $indent, {} );
 
 		} else {
-			die "CODE INCOMPLETE";
+			Carp::croak("Cannot serialize " . ref($cursor));
 		}
 	}
 
@@ -325,35 +353,49 @@ sub _write_scalar {
 		$str =~ s/\\/\\\\/g;
 		$str =~ s/"/\\"/g;
 		$str =~ s/\n/\\n/g;
-		$str =~ s/([\x00-\x1f])/\\$UNPRINTABLE[ord($1)]/ge;
+		$str =~ s/([\x00-\x1f])/\\$UNPRINTABLE[ord($1)]/g;
 		return qq{"$str"};
 	}
 	if ( length($str) == 0 or $str =~ /\s/ ) {
-		$str =~ s/'/''/;
+		$str =~ s/\'/\'\'/;
 		return "'$str'";
 	}
 	return $str;
 }
 
 sub _write_array {
-	my ($self, $indent, $array) = @_;
+	my ($self, $array, $indent, $seen) = @_;
+	if ( $seen->{refaddr($array)}++ ) {
+		die "YAML::Tiny does not support circular references";
+	}
 	my @lines  = ();
 	foreach my $el ( @$array ) {
 		my $line = ('  ' x $indent) . '-';
-		if ( ! ref $el ) {
+		my $type = ref $el;
+		if ( ! $type ) {
 			$line .= ' ' . $self->_write_scalar( $el );
 			push @lines, $line;
 
-		} elsif ( ref $el eq 'ARRAY' ) {
-			push @lines, $line;
-			push @lines, $self->_write_array( $indent + 1, $el );
+		} elsif ( $type eq 'ARRAY' ) {
+			if ( @$el ) {
+				push @lines, $line;
+				push @lines, $self->_write_array( $el, $indent + 1, $seen );
+			} else {
+				$line .= ' []';
+				push @lines, $line;
+			}
 
-		} elsif ( ref $el eq 'HASH' ) {
-			push @lines, $line;
-			push @lines, $self->_write_hash( $indent + 1, $el );
+		} elsif ( $type eq 'HASH' ) {
+			if ( keys %$el ) {
+				push @lines, $line;
+				push @lines, $self->_write_hash( $el, $indent + 1, $seen );
+			} else {
+				$line .= ' {}';
+				push @lines, $line;
+			}
 
 		} else {
-			die "CODE INCOMPLETE";
+			die "YAML::Tiny does not support $type references";
 		}
 	}
 
@@ -361,25 +403,39 @@ sub _write_array {
 }
 
 sub _write_hash {
-	my ($self, $indent, $hash) = @_;
+	my ($self, $hash, $indent, $seen) = @_;
+	if ( $seen->{refaddr($hash)}++ ) {
+		die "YAML::Tiny does not support circular references";
+	}
 	my @lines  = ();
 	foreach my $name ( sort keys %$hash ) {
 		my $el   = $hash->{$name};
 		my $line = ('  ' x $indent) . "$name:";
-		if ( ! ref $el ) {
+		my $type = ref $el;
+		if ( ! $type ) {
 			$line .= ' ' . $self->_write_scalar( $el );
 			push @lines, $line;
 
-		} elsif ( ref $el eq 'ARRAY' ) {
-			push @lines, $line;
-			push @lines, $self->_write_array( $indent + 1, $el );
+		} elsif ( $type eq 'ARRAY' ) {
+			if ( @$el ) {
+				push @lines, $line;
+				push @lines, $self->_write_array( $el, $indent + 1, $seen );
+			} else {
+				$line .= ' []';
+				push @lines, $line;
+			}
 
-		} elsif ( ref $el eq 'HASH' ) {
-			push @lines, $line;
-			push @lines, $self->_write_hash( $indent + 1, $el );
+		} elsif ( $type eq 'HASH' ) {
+			if ( keys %$el ) {
+				push @lines, $line;
+				push @lines, $self->_write_hash( $el, $indent + 1, $seen );
+			} else {
+				$line .= ' {}';
+				push @lines, $line;
+			}
 
 		} else {
-			die "CODE INCOMPLETE";
+			die "YAML::Tiny does not support $type references";
 		}
 	}
 
@@ -388,13 +444,13 @@ sub _write_hash {
 
 # Set error
 sub _error {
-	$errstr = $ERROR{$_[1]} ? "$ERROR{$_[1]} ($_[1])" : $_[1];
+	$YAML::Tiny::errstr = $_[1];
 	undef;
 }
 
 # Retrieve error
 sub errstr {
-	$errstr;
+	$YAML::Tiny::errstr;
 }
 
 
@@ -404,59 +460,64 @@ sub errstr {
 #####################################################################
 # YAML Compatibility
 
-use vars qw{@EXPORT_OK};
-BEGIN {
-	require Exporter;
-	@EXPORT_OK = qw{ Load Dump };
-}
-
-=pod
-
-=head1 FUNCTIONS
-
-YAML::Tiny implements two functions to add compatibility with the L<YAML>
-API. These should be a drop-in replacement, except that YAML::Tiny will
-B<not> export functions by default, and so you will need to explicitly
-import the functions.
-
-=head2 Dump
-
-  my $string = Dump(list-of-Perl-data-structures);
-
-Turn Perl data into YAML. This function works very much like Data::Dumper::Dumper().
-
-It takes a list of Perl data strucures and dumps them into a serialized form.
-
-It returns a string containing the YAML stream.
-
-The structures can be references or plain scalars.
-
-=cut
-
 sub Dump {
 	YAML::Tiny->new(@_)->write_string;
 }
 
-=pod
-
-=head2 Load
-
-  my @documents = Load(string-containing-a-YAML-stream);
-
-Turn YAML into Perl data. This is the opposite of Dump.
-
-Just like Storable's thaw() function or the eval() function in relation to Data::Dumper.
-
-It parses a string containing a valid YAML stream into a list of Perl data structures.
-
-=cut
-
 sub Load {
 	my $self = YAML::Tiny->read_string(@_)
 		or Carp::croak("Failed to load YAML document from string");
 	return @$self;	
 }
 
+BEGIN {
+	*freeze = *Dump;
+	*thaw   = *Load;
+}
+
+sub DumpFile {
+	my $file = shift;
+	YAML::Tiny->new(@_)->write($file);
+}
+
+sub LoadFile {
+	my $self = YAML::Tiny->read($_[0])
+		or Carp::croak("Failed to load YAML document from '" . ($_[0] || '') . "'");
+	return @$self;
+}
+
+
+
+
+
+#####################################################################
+# Use Scalar::Util if possible, otherwise emulate it
+
+BEGIN {
+	eval {
+		require Scalar::Util;
+	};
+	if ( $@ ) {
+		# Failed to load Scalar::Util
+		eval <<'END_PERL';
+sub refaddr {
+	my $pkg = ref($_[0]) or return undef;
+	if (!!UNIVERSAL::can($_[0], 'can')) {
+		bless $_[0], 'Scalar::Util::Fake';
+	} else {
+		$pkg = undef;
+	}
+	"$_[0]" =~ /0x(\w+)/;
+	my $i = do { local $^W; hex $1 };
+	bless $_[0], $pkg if defined $pkg;
+	$i;
+}
+END_PERL
+	} else {
+		Scalar::Util->import('refaddr');
+	}
+}
+
 1;
 
 __END__
@@ -469,43 +530,21 @@ YAML::Tiny - Read/Write YAML files with as little code as possible
 
 =head1 PREAMBLE
 
-B<WARNING: THIS MODULES IS HIGHLY EXPERIMENTAL AND SUBJECT TO CHANGE>
-
-The YAML specification is huge. Like, B<really> huge. It contains all the
+The YAML specification is huge. Really, B<really> huge. It contains all the
 functionality of XML, except with flexibility and choice, which makes it
-easier to read, but with a full specification that is more complex than XML.
+easier to read, but with a formal specification that is more complex than
+XML.
 
-The pure-Perl implementation L<YAML> costs just over 4 megabytes of memory
-to load. Just like with Windows .ini files (3 meg to load) and CSS (3.5 meg
-to load) the situation is just asking for a B<YAML::Tiny> module, an
+The original pure-Perl implementation L<YAML> costs just over 4 megabytes of
+memory to load. Just like with Windows .ini files (3 meg to load) and CSS
+(3.5 meg to load) the situation is just asking for a B<YAML::Tiny> module, an
 incomplete but correct and usable subset of the functionality, in as little
 code as possible.
 
-Now, given the YAML features one would need in order to have something
-that is usable for things like META.yml and simple configuration files,
-there is still enough complexity that I'm not sure if it is even possible
-to do a YAML::Tiny module.
-
-So I'm going to impose some ground rules before I start.
-
 Like the other C<::Tiny> modules, YAML::Tiny will have no non-core
 dependencies, not require a compiler, and be back-compatible to at least
 perl 5.005_03, and ideally 5.004.
 
-And I'm setting a hard-limit of 400k of memory to load it
-(1/10th of YAML.pm).
-
-I plan to implement features starting at the most common and working
-towards the least common, but if we hit 400k limit then we stop until
-we can find a way to squish the same functionality into less code and
-free some up.
-
-At this point, other than unquoted scalars, arrays, hashes and ASCII,
-I promise nothing.
-
-So do B<not> use this module for anything other than experimentation.
-It's only just getting started.
-
 =head1 SYNOPSIS
 
     #############################################
@@ -550,18 +589,20 @@ It's only just getting started.
 
 =head1 DESCRIPTION
 
-B<YAML::Tiny> is a perl class to read and write YAML-style files with as
-little code as possible, reducing load time and memory overhead.
+B<YAML::Tiny> is a perl class for reading and writing YAML-style files,
+written with as little code as possible, reducing load time and memory
+overhead.
 
 Most of the time it is accepted that Perl applications use a lot
 of memory and modules. The B<::Tiny> family of modules is specifically
 intended to provide an ultralight and zero-dependency alternative to
-the standard modules.
+many more-thorough standard modules.
 
-This module is primarily for reading human-written files (like config files)
-and generating very simple human-readable files. Note that I said
-B<human-readable> and not B<geek-readable>. The sort of files that your
-average manager or secretary should be able to look at and make sense of.
+This module is primarily for reading human-written files (like simple
+config files) and generating very simple human-readable files. Note that
+I said B<human-readable> and not B<geek-readable>. The sort of files that
+your average manager or secretary should be able to look at and make
+sense of.
 
 L<YAML::Tiny> does not generate comments, it won't necesarily preserve the
 order of your hashes, and it will normalise if reading in and writing out
@@ -570,7 +611,7 @@ again.
 It only supports a very basic subset of the full YAML specification.
 
 Usage is targetted at files like Perl's META.yml, for which a small and
-easily-embeddable module would be highly useful.
+easily-embeddable module is extremely attractive.
 
 Features will only be added if they are human readable, and can be written
 in a few lines of code. Please don't be offended if your request is
@@ -625,6 +666,51 @@ Generates the file content for the object and returns it as a string.
 When an error occurs, you can retrieve the error message either from the
 C<$YAML::Tiny::errstr> variable, or using the C<errstr()> method.
 
+=head1 FUNCTIONS
+
+YAML::Tiny implements a number of functions to add compatibility with
+the L<YAML> API. These should be a drop-in replacement, except that
+YAML::Tiny will B<not> export functions by default, and so you will need
+to explicitly import the functions.
+
+=head2 Dump
+
+  my $string = Dump(list-of-Perl-data-structures);
+
+Turn Perl data into YAML. This function works very much like Data::Dumper::Dumper().
+
+It takes a list of Perl data strucures and dumps them into a serialized form.
+
+It returns a string containing the YAML stream.
+
+The structures can be references or plain scalars.
+
+=head2 Load
+
+  my @documents = Load(string-containing-a-YAML-stream);
+
+Turn YAML into Perl data. This is the opposite of Dump.
+
+Just like L<Storable>'s thaw() function or the eval() function in relation
+to L<Data::Dumper>.
+
+It parses a string containing a valid YAML stream into a list of Perl data
+structures.
+
+=head2 freeze() and thaw()
+
+Aliases to Dump() and Load() for L<Storable> fans. This will also allow
+YAML::Tiny to be plugged directly into modules like POE.pm, that use the
+freeze/thaw API for internal serialization.
+
+=head2 DumpFile(filepath, list)
+
+Writes the YAML stream to a file instead of just returning a string.
+
+=head2 LoadFile(filepath)
+
+Reads the YAML stream from a file instead of a string.
+
 =head1 SUPPORT
 
 Bugs should be reported via the CPAN bug tracker at
@@ -640,16 +726,16 @@ For other issues, or commercial enhancement or support, please contact
 
 =head1 AUTHOR
 
-Adam Kennedy E<lt>cpan@ali.asE<gt>
+Adam Kennedy E<lt>adamk@cpan.orgE<gt>
 
 =head1 SEE ALSO
 
 L<YAML>, L<YAML::Syck>, L<Config::Tiny>, L<CSS::Tiny>,
-L<http://use.perl.org/~Alias/journal/29427>
+L<http://use.perl.org/~Alias/journal/29427>, L<http://ali.as/>
 
 =head1 COPYRIGHT
 
-Copyright 2006 Adam Kennedy. All rights reserved.
+Copyright 2006 - 2007 Adam Kennedy.
 
 This program is free software; you can redistribute
 it and/or modify it under the same terms as Perl itself.
